#include <linux/module.h>
#include <linux/cdev.h>
#include <linux/interrupt.h>
#include <linux/gpio/consumer.h>
#include <linux/of_device.h>

static dev_t myhw_dev_number;
static struct cdev driver_object;
static struct class *myhw_class;
static struct device *myhw_dev;

static struct gpio_desc *gpio_irq, *gpio_trigger;
static int myhw_irq;

static irqreturn_t hard_isr( int irq, void *data )
{
	return IRQ_WAKE_THREAD;
}

static irqreturn_t isr_thread( int irq, void *data )
{
	pr_info("isr_thread( %d, %p )\n", irq, data );
	return IRQ_HANDLED;
}

static int config_gpios( void )
{
	int err;
	struct device_node *nodeptr = myhw_dev->of_node;

        nodeptr = of_find_node_by_name(nodeptr, "myhw");
	myhw_dev->of_node = nodeptr;

        pr_info("config_gpios( %p, %p )\n", myhw_dev, nodeptr);

	gpio_trigger = gpiod_get( myhw_dev, "my-trigger", GPIOD_OUT_LOW );
	if (IS_ERR(gpio_trigger)) {
		pr_err("gpiod_get failed with my-trigger\n");
		return -EIO;
	}
	pr_info("gpio-trigger configured...");

	gpio_irq = gpiod_get( myhw_dev, "my-irq", GPIOD_IN );
	if (IS_ERR(gpio_irq)) {
		pr_err("gpiod_get failed with my-irq\n");
		return -EIO;
	}
	myhw_irq = gpiod_to_irq( gpio_irq );
	dev_info(myhw_dev, "using irq: %d\n", myhw_irq);

	err = request_threaded_irq( myhw_irq, hard_isr,isr_thread,
		IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING,
		"myhw", &driver_object);
	if (err) {
		pr_err("request_irq failed with %d\n",err);
		gpiod_put( gpio_irq );
		gpiod_put( gpio_trigger );
		return -EIO;
	}
	return 0;
}

static int free_gpios( void )
{
	if (gpio_irq)
		gpiod_put( gpio_irq );
	if (gpio_trigger)
		gpiod_put( gpio_trigger );
	free_irq(myhw_irq, &driver_object);
	return 0;
}

static ssize_t device_write( struct file *instanz, const char __user *user,
		size_t count, loff_t *offset )
{
	unsigned long not_copied=0, to_copy=0;
	int value=0;

	to_copy = min( count, sizeof(value) );
	not_copied=copy_from_user(&value,user,to_copy);
	pr_info("device_write( %d )", value);
	gpiod_set_value( gpio_trigger, value );
	*offset += to_copy-not_copied;
	return to_copy-not_copied;
}

static ssize_t device_read( struct file *instanz, char __user *user,
		size_t count, loff_t *offset )
{
	unsigned long not_copied, to_copy;
	int value;

	value = gpiod_get_value( gpio_irq );

	to_copy = min( count, sizeof(value) );
	not_copied=copy_to_user(user,&value,to_copy);
	*offset += to_copy-not_copied;
	return to_copy-not_copied;
}

static struct file_operations fops = {
	.owner   = THIS_MODULE,
	.read    = device_read,
	.write   = device_write,
};

static int __init mod_init( void )
{
	if (alloc_chrdev_region(&myhw_dev_number,0,1,"myhw")<0)
		return -EIO;
	cdev_init( &driver_object, &fops );
	driver_object.owner = THIS_MODULE;

	if (cdev_add(&driver_object,myhw_dev_number,1))
		goto free_devnum;
	myhw_class = class_create( "myhw" );
	if (IS_ERR( myhw_class )) {
		pr_err( "myhw: no udev support\n");
		goto free_cdev;
	}
	myhw_dev = device_create(myhw_class,NULL,myhw_dev_number,
		NULL, "%s", "myhw" );
	if (IS_ERR( myhw_dev )) {
		pr_err( "myhw: device_create failed\n");
		goto free_class;
	}
	if ( config_gpios() )
		goto free_device;
	return 0;
free_device:
	device_destroy( myhw_class, myhw_dev_number );
free_class:
	class_destroy( myhw_class );
free_cdev:
	cdev_del( &driver_object );
free_devnum:
	unregister_chrdev_region( myhw_dev_number, 1 );
	return -EIO;
}

static void __exit mod_exit( void )
{
	free_gpios();
	device_destroy( myhw_class, myhw_dev_number );
	class_destroy( myhw_class );
	cdev_del( &driver_object );
	unregister_chrdev_region( myhw_dev_number, 1 );
	return;
}

module_init( mod_init );
module_exit( mod_exit );

MODULE_AUTHOR("Juergen Quade");
MODULE_LICENSE("GPL");
